<?php

namespace Drupal\facets_range_dropdowns\Plugin\facets\processor;

use Drupal\facets\FacetInterface;
use Drupal\facets\Processor\BuildProcessorInterface;
use Drupal\facets\Processor\PostQueryProcessorInterface;
use Drupal\facets\Processor\PreQueryProcessorInterface;
use Drupal\facets\Processor\ProcessorPluginBase;
use Drupal\facets_range_dropdowns\BoundResultInterface;
use Drupal\facets_range_dropdowns\Result\BoundResult;

/**
 * Class RangeProcessor.
 *
 * @package Drupal\facets_range_dropdowns\Plugin\facets\processor
 *
 * @FacetsProcessor(
 *   id = "range",
 *   label = @Translation("Range"),
 *   description = @Translation("Add range results for all the steps between min and max."),
 *   stages = {
 *     "pre_query" = 60,
 *     "post_query" = 60,
 *     "build" = 20
 *   }
 * )
 */
class RangeProcessor extends ProcessorPluginBase implements PreQueryProcessorInterface, PostQueryProcessorInterface, BuildProcessorInterface {

  /**
   * {@inheritdoc}
   */
  public function preQuery(FacetInterface $facet) {
    $activeItems = $facet->getActiveItems();

    array_walk($activeItems, function (&$item) {
      if (preg_match('/\(min:((?:-)?[\d\.]+),max:((?:-)?[\d\.]+)\)/i', $item, $matches)) {
        $item = [$matches[1], $matches[2]];
      }
      else {
        $item = NULL;
      }
    });
    $facet->setActiveItems($activeItems);
  }

  /**
   * {@inheritdoc}
   */
  public function postQuery(FacetInterface $facet) {
    $config = $facet->getWidgetInstance()->getConfiguration();

    // Get an array with the sorted results.
    $realResults = $this->getSortedResults($facet->getResults());

    // Determine step size and min and max.
    $step = $config['step'];
    [$min, $max] = $this->getMinMax($facet, $realResults);

    // Get the active items set via preQuery().
    $activeMin = $facet->getActiveItems()[0][0] ?? NULL;
    $activeMax = $facet->getActiveItems()[0][1] ?? NULL;

    // Create an array of all results between min and max by the step from the
    // configuration.
    $facetResults = [];
    for ($i = $min; $i < $max + $step; $i += $step) {
      if ($i > $max) {
        $i = $max;
      }

      $displayValue = (string) $i;
      if (isset($config['prefix'])) {
        $displayValue = $config['prefix'] . $displayValue;
      }
      if (isset($config['suffix'])) {
        $displayValue = $displayValue . $config['suffix'];
      }

      $lowerCount = $this->getTotalCountWithinBounds($facet->getResults(), $i, $activeMax);
      $facetResults[] = new BoundResult($facet, (float) $i, $displayValue, max(0, $lowerCount), 'lower');
      $upperCount = $this->getTotalCountWithinBounds($facet->getResults(), $activeMin, $i);
      $facetResults[] = new BoundResult($facet, (float) $i, $displayValue, max(0, $upperCount), 'upper');
    }

    // Overwrite the current facet values with the generated results.
    $facet->setResults($facetResults);
  }

  /**
   * {@inheritdoc}
   */
  public function build(FacetInterface $facet, array $results) {
    // Determine the active filters by inspecting the URL processor.
    $activeFilters = $facet->getProcessors()['url_processor_handler']
      ->getProcessor()
      ->getActiveFilters();

    if (isset($activeFilters[''])) {
      unset($activeFilters['']);
    }

    // Determine min and max.
    [$min, $max] = $this->getMinMax($facet, $this->getSortedResults($results));

    // Active items set via preQuery().
    $activeMin = $facet->getActiveItems()[0][0] ?? NULL;
    $activeMax = $facet->getActiveItems()[0][1] ?? NULL;

    /** @var \Drupal\facets\Result\ResultInterface[] $results */
    foreach ($results as &$result) {
      $newActiveFilters = $activeFilters;
      unset($newActiveFilters[$facet->id()]);

      // Default min and max values.
      $minValue = $min;
      $maxValue = $max;

      if ($result instanceof BoundResultInterface) {
        switch ($result->getBound()) {
          case 'lower':
            // For a result intended for lower bound, the minimum is the actual
            // value.
            $minValue = $result->getRawValue();
            if (isset($activeMin) && $minValue == $activeMin) {
              $result->setActiveState(TRUE);
            }
            // Preserve current active max.
            if (isset($activeMax)) {
              $maxValue = $activeMax;
            }
            break;

          case 'upper':
            // For a result intended for upper bound, the maximum is the actual
            // value.
            $maxValue = $result->getRawValue();
            if (isset($activeMax) && $maxValue == $activeMax) {
              $result->setActiveState(TRUE);
            }
            // Preserve current active min.
            if (isset($activeMin)) {
              $minValue = $activeMin;
            }
            break;

          default:
            break;
        }
      }

      $newActiveFilters[$facet->id()][] = "(min:$minValue,max:$maxValue)";
      $url = \Drupal::service('facets.utility.url_generator')->getUrl($newActiveFilters, FALSE);
      $result->setUrl($url);
    }

    return $results;
  }

  /**
   * Gets the results from a facet as a sorted array.
   *
   * @param \Drupal\facets\Result\ResultInterface[] $results
   *   Facet.
   *
   * @return array
   *   Array with sorted results.
   */
  protected function getSortedResults(array $results) {
    $realResults = [];

    // Determine the actual results.
    foreach ($results as $result) {
      $realResults['f_' . (float) $result->getRawValue()] = [
        'value' => (float) $result->getRawValue(),
        'count' => $result->getCount(),
      ];
    }

    // Sort the results by value.
    uasort($realResults, function ($a, $b) {
      if ($a['value'] === $b['value']) {
        return 0;
      }

      return $a['value'] < $b['value'] ? -1 : 1;
    });

    return $realResults;
  }

  /**
   * Gets the min and max values based on the facet configuration and results.
   *
   * @param \Drupal\facets\FacetInterface $facet
   *   Facet.
   * @param array $results
   *   Results.
   *
   * @return array
   *   Indexed array with min and max values respectively.
   */
  protected function getMinMax(FacetInterface $facet, array $results) {
    $config = $facet->getWidgetInstance()->getConfiguration();

    $step = $config['step'];
    if ($config['min_type'] === 'fixed' && $config['max_type'] !== 'fixed') {
      $min = $config['min_value'];
      $max = end($results)['value'] ?? $min;
      // If max is not divisible by step, we should add the remainder to max to
      // make sure that we do not lose any possible values.
      if (fmod($max, $step) != 0) {
        $max = $max + ($step - fmod($max, $step));
      }
    }
    elseif ($config['min_type'] !== 'fixed' && $config['max_type'] === 'fixed') {
      $min = reset($results)['value'] ?? 0;
      $max = $config['max_value'];
    }
    elseif ($config['min_type'] === 'fixed' && $config['max_type'] === 'fixed') {
      $min = $config['min_value'];
      $max = $config['max_value'];
    }
    else {
      $min = reset($results)['value'] ?? 0;
      $max = end($results)['value'] ?? $min;
      // If max is not divisible by step, we should add the remainder to max to
      // make sure that we do not lose any possible values.
      if (fmod($max, $step) != 0) {
        $max = $max + ($step - fmod($max, $step));
      }
    }

    return [$min, $max];
  }

  /**
   * Returns the total number of search results for the given facet results.
   *
   * @param \Drupal\facets\Result\ResultInterface[] $results
   *   Facet results.
   * @param int|null $lowerBound
   *   Lower bound.
   * @param int|null $upperBound
   *   Upper bound.
   *
   * @return int
   *   Total number of search results.
   */
  protected function getTotalCountWithinBounds(array $results, $lowerBound = NULL, $upperBound = NULL) {
    $count = 0;
    foreach ($results as $result) {
      if ($lowerBound !== NULL && $result->getRawValue() < $lowerBound) {
        continue;
      }
      if ($upperBound !== NULL && $result->getRawValue() > $upperBound) {
        continue;
      }
      $count += $result->getCount();
    }

    return $count;
  }

}
